using System;
using System.Net.Sockets;
using System.Threading;
using System.Text;
using System.IO;
using System.Xml;
using System.Data.Odbc;
using System.Globalization;
namespace TCPService
{
	/// <summary>
	/// Summary description for TCPSocketListener.
	/// </summary>
	public class TCPSocketListener
	{

	
		public enum STATE{LOGIN, ACTIVE, UNACTIVE};

		/// <summary>
		/// Variables that are accessed by other classes indirectly.
		/// </summary>
		private Socket m_clientSocket = null;
		private bool m_stopClient=false;
		private Thread m_clientListenerThread=null;
		private bool m_markedForDeletion=false;
                private string m_user;
                private OdbcConnection m_conn;

		/// <summary>
		/// Working Variables.
		/// </summary>
		private StringBuilder m_oneLineBuf=new StringBuilder();
		private STATE m_processState=STATE.LOGIN;

		private DateTime m_lastReceiveDateTime;
		private DateTime m_currentReceiveDateTime;
		
		/// <summary>
		/// Client Socket Listener Constructor.
		/// </summary>
		/// <param name="clientSocket"></param>
		public TCPSocketListener(Socket clientSocket)
		{
			m_clientSocket = clientSocket;
                        m_conn=new OdbcConnection("DSN=Message");
                        m_conn.Open();
		}

		/// <summary>
		/// Client SocketListener Destructor.
		/// </summary>
		~TCPSocketListener()
		{
                	m_conn.Close();
			StopSocketListener();
		}

		/// <summary>
		/// Method that starts SocketListener Thread.
		/// </summary>
		public void StartSocketListener()
		{
			if (m_clientSocket!= null)
			{
				m_clientListenerThread = 
					new Thread(new ThreadStart(SocketListenerThreadStart));

				m_clientListenerThread.Start();
			}
		}

		/// <summary>
		/// Thread method that does the communication to the client. This 
		/// thread tries to receive from client and if client sends any data
		/// then parses it and again wait for the client data to come in a
		/// loop. The recieve is an indefinite time receive.
		/// </summary>
		private void SocketListenerThreadStart()
		{
			int size=0;
			Byte [] byteBuffer = new Byte[1024];

			m_lastReceiveDateTime = DateTime.Now;
			m_currentReceiveDateTime = DateTime.Now;

			Timer t= new Timer(new TimerCallback(CheckClientCommInterval),
				null,15000,15000);
                        Timer t1 = new Timer(new TimerCallback(CheckMessageBuffer),null,200,200);
                        Timer t2 = new Timer(new TimerCallback(CheckUserOnline),null,10000,10000);
                        Timer t3 = new Timer(new TimerCallback(CheckFriendInvites),null,1000,1000);
			while (!m_stopClient)
			{
				try
				{
                                        while(m_clientSocket.Available==0);
					size = m_clientSocket.Receive(byteBuffer,m_clientSocket.Available,0);

					m_currentReceiveDateTime=DateTime.Now;
					ParseReceiveBuffer(byteBuffer, size);
				}
				catch (SocketException se)
				{
					m_stopClient=true;
					m_markedForDeletion=true;
				}
			}
			t.Change(Timeout.Infinite, Timeout.Infinite);
			t=null;
                        t1.Change(Timeout.Infinite,Timeout.Infinite);
                        t1=null;
                        t2.Change(Timeout.Infinite,Timeout.Infinite);
                        t2=null;
                        t3.Change(Timeout.Infinite,Timeout.Infinite);
                        t3=null;
		}

		/// <summary>
		/// Method that stops Client SocketListening Thread.
		/// </summary>
		public void StopSocketListener()
		{
			if (m_clientSocket!= null)
			{
				m_stopClient=true;
				m_clientSocket.Close();

				// Wait for one second for the the thread to stop.
				m_clientListenerThread.Join(1000);

				// If still alive; Get rid of the thread.
				if (m_clientListenerThread.IsAlive)
				{
					m_clientListenerThread.Abort();
				}
				m_clientListenerThread=null;
				m_clientSocket=null;
				m_markedForDeletion=true;
			}
		}

		/// <summary>
		/// Method that returns the state of this object i.e. whether this
		/// object is marked for deletion or not.
		/// </summary>
		/// <returns></returns>
		public bool IsMarkedForDeletion()
		{
			return m_markedForDeletion;
		}

		/// <summary>
		/// This method parses data that is sent by a client using TCP/IP.
		/// As per the "Protocol" between client and this Listener, client
		/// sends each line of information by appending "CRLF" (Carriage Return
		/// and Line Feed). But since the data is transmitted from client to
		/// here by TCP/IP protocol, it is not guarenteed that each line that
		/// arrives ends with a "CRLF". So the job of this method is to make a
		/// complete line of information that ends with "CRLF" from the data
		/// that comes from the client and get it processed.
		/// </summary>
		/// <param name="byteBuffer"></param>
		/// <param name="size"></param>
		private void ParseReceiveBuffer(Byte [] byteBuffer, int size)
		{
			string data = Encoding.ASCII.GetString(byteBuffer,0, size);
			int lineEndIndex = 0;

			// Check whether data from client has more than one line of
			// information, where each line of information ends with "CRLF"
			// ("\r\n"). If so break data into different lines and process
			// separately.
			do
			{
				lineEndIndex =	data.IndexOf("\r\n");
				if(lineEndIndex != -1)
				{
					m_oneLineBuf=m_oneLineBuf.Append(data,0,lineEndIndex+2);
					ProcessClientData(m_oneLineBuf.ToString());
					m_oneLineBuf.Remove(0,m_oneLineBuf.Length);
					data = data.Substring(lineEndIndex+2,
						data.Length -lineEndIndex-2);
				}
				else
				{
					// Just append to the existing buffer.
					m_oneLineBuf=m_oneLineBuf.Append(data);
				}
			}while(lineEndIndex != -1);
		}

		/// <summary>
		/// Method that Process the client data as per the protocol.
		/// The protocol works like this.
		/// 1. Client first send a file name that ends with "CRLF".
		///
		/// 2. This SocketListener has to return the length of the file name
		/// to the client for validation. If the length matches with the length
		/// what  client had sent earlier, then client starts sending lines of
		/// information. Otherwise socket will be closed by the client.
		///
		/// 3. Each line of information that client sends will end with "CRLF".
		///
		/// 4. This socketListener has to store each line of information in
		/// a text file whoose file name has been sent by the client earlier.
		///
		/// 5. As a last line of information client sends "[EOF]" line which
		/// also ends with "CRLF" ("\r\n"). This signals this SocketListener
		/// for an end of file and intern this SocketListener sends the total
		/// length of the data (lines of information excludes file name that
		/// was sent earlier) back to client for validation.
		/// </summary>
		/// <param name="oneLine"></param>
		private void ProcessClientData(String oneLine)
		{
			OdbcCommand cmd;
			ASCIIEncoding ASCII = new ASCIIEncoding();
			XmlDocument doc= new XmlDocument();
			doc.LoadXml(oneLine);
			XmlAttributeCollection list;
			XmlAttribute attr;
			int i;
                        string logClientOut=@"<LOGGEDOUT user="""+m_user+@"""/>"+"\r\n";
                        byte[] bytesSent;
                        bytesSent=ASCII.GetBytes(oneLine);
                        m_clientSocket.Send(bytesSent);
                        OdbcDataReader data;
			switch(m_processState)
			{
				case STATE.LOGIN:
                                lock(this){

					string user="";
					string pass="";
                                        user=doc.DocumentElement.GetAttribute("user");
					pass=doc.DocumentElement.GetAttribute("pass");
                                        string failed=@"<LOGIN user="""+user+@"""result=""0""/>"+"\r\n";
					string success=@"<LOGIN user="""+user+@"""result=""1""/>"+"\r\n";
                                        byte[] bytes;
					if(user.Equals("")||pass.Equals("")){
						m_stopClient=true;
						bytes=ASCII.GetBytes(failed);
						m_clientSocket.Send(bytes);
						return;
					}
					cmd=new OdbcCommand(@"SELECT * FROM Authentication, Banned WHERE username='"+user+"' AND password='"+pass+"' AND username NOT IN (SELECT username FROM online);",m_conn);
					data=cmd.ExecuteReader();
					if(!data.HasRows){
						m_stopClient=true;
						bytes=ASCII.GetBytes(failed);
                                                System.Diagnostics.Debug.WriteLine(ASCII.GetString(bytes));
						m_clientSocket.Send(bytes);
						return;
					}
                                        if(data.GetBoolean(1))
                                        	bytes=ASCII.GetBytes(@"<LOGIN user="""+user+@"""result=""2""/>"+"\r\n");
                                        else
                                        	bytes=ASCII.GetBytes(success);
                                        m_clientSocket.Send(bytes);
                                        System.Diagnostics.Debug.WriteLine(ASCII.GetString(bytes));
                                        if(data.GetBoolean(1)){
                                        	data.Close();
                                                break;
                                        }

					data.Close();
                                        DateTime dt=DateTime.Now;
					cmd=new OdbcCommand("INSERT INTO Online VALUES('"+user+"',#"+dt.ToString("MM/dd/yyyy")+"#,#"+dt.ToString("MM/dd/yyyy")+"#,2);",m_conn);
                                        try{
					cmd.ExecuteNonQuery();
                                        }
                                        catch (OdbcException oe){
						int count=0;
                                        	for(;count<oe.Errors.Count;count++)
                                                	System.Diagnostics.Debug.WriteLine(oe.Errors[count].Message);
                                        }
					m_user=user;
					m_processState=STATE.ACTIVE;
                                        }
					break;
				case STATE.ACTIVE:
                                        lock(this){
                                        switch(doc.DocumentElement.Name.ToUpper()){
						case "LOGOUT":
							string send="";
							list =doc.DocumentElement.Attributes;
							attr = list[0];
							if(attr.Name.ToUpper().Equals("USER") && attr.Value.Equals(m_user)){
								//Delete user from 'Online'
								cmd = new OdbcCommand("DELETE FROM Online WHERE username='"+m_user+"';",m_conn);
								cmd.ExecuteNonQuery();
								//Change State to UNACTIVE
								m_processState=STATE.UNACTIVE;
								//Stop Process
								m_stopClient=true;
								//Set positive Response to client
								send=@"<LOGOUT result=""1""/>"+"\r\n";
							}
							else {
								//Set Negative Response to Client
								send=@"<LOGOUT result=""0""/>"+"\r\n";
							}
							//Send response to client
							bytesSent=ASCII.GetBytes(send);
							m_clientSocket.Send(bytesSent);
							break;
						case "MESSAGE":
							list =doc.DocumentElement.Attributes;
							string result;
							string from="";
							string to="";
							for (i=0;i<list.Count;i++){
								attr = list[i];
								switch(attr.Name.ToUpper()){
									case "FROM":
										from = attr.Value;
										break;
									case "TO":
										to = attr.Value;
										break;
								}
							}
							if (!from.Equals(m_user))
								result="0";
							else {
								cmd=new OdbcCommand("SELECT * FROM Authentication WHERE username='"+to+"';",m_conn);
								data = cmd.ExecuteReader();
								if(!data.HasRows){
									data.Close();
									result="0";
								}
								else {
									data.Close();
									cmd = new OdbcCommand("SELECT * Online WHERE username='"+to+"';",m_conn);
									data = cmd.ExecuteReader();
									if(data.HasRows){
										data.Read();
										result="1";
									}
									else
										result="-1";
									data.Close();
									cmd = new OdbcCommand("INSERT INTO Buffer VALUES ("+doc.DocumentElement.Value+","+from+","+to+");",m_conn);
                                                                        try{cmd.ExecuteNonQuery();}
                                                                        catch(OdbcException oe){
                                                                        	result="0";
                                                                        }									cmd.ExecuteNonQuery();
								}
							}
							string resp =@"<MESSAGE result="" "+result+@" ""/>"+"\r\n";
							bytesSent=ASCII.GetBytes(resp);
							m_clientSocket.Send(bytesSent);
							break;
						case "STATUS":
							string name=doc.DocumentElement.GetAttribute("user");
                                                        string type=doc.DocumentElement.GetAttribute("type");
                                                        string res="";
                                                        switch (type.ToUpper()){
                                                                case "SET":
								if(m_user.Equals(name)){
									cmd = new OdbcCommand("UPDATE Online SET Status="+doc.DocumentElement.GetAttribute("code"),m_conn);

									res=@"<STATUS type=""set"" result="""+cmd.ExecuteNonQuery().ToString()+@"""/>"+"\r\n";
								}
								else
									res=@"<STATUS type=""set"" result=""0""/>"+"\r\n";
                                                                break;
                                                                case "GET":
                                                                if(m_user.Equals(name)){
                                                                	cmd = new OdbcCommand("SELECT Status FROM Online WHERE Username = '"+name+"';",m_conn);
                                                                        data=cmd.ExecuteReader();
                                                                        if(data.HasRows){
                                                                        	res=@"<STATUS type=""get"" result="""+data.GetInt32(0)+@"""/>"+"\r\n";
                                                                        }
                                                                        else
                                                                        	res=@"<STATUS type=""get"" result=""-1""/>"+"\r\n";
                                                                        data.Close();
                                                                }
                                                                break;
                                                                case "LIST":
                                                                cmd = new OdbcCommand("SELECT * FROM Status;",m_conn);
                                                                data = cmd.ExecuteReader();
                                                                res="<STATUSLIST>";
                                                                while(data.Read()){
                                                                	res+=@"<STATUS id="""+data.GetInt32(0)+@""">"+data.GetString(1)+"</STATUS>";
                                                                }
                                                                res+="</STATUSLIST>\r\n";
                                                                break;
                                                        }
                                                        bytesSent=ASCII.GetBytes(res);
							m_clientSocket.Send(bytesSent);
                                                        break;


						case "FRIEND":
							type=doc.DocumentElement.GetAttribute("type");
							string friend;
							string rep;
							switch(type.ToUpper()){
							  case "ADD":
								if(doc.DocumentElement.GetAttribute("user").Equals(m_user)){
									friend=doc.DocumentElement.GetAttribute("friend");
									cmd = new OdbcCommand("SELECT username FROM Authentication WHERE username = '"+friend+"';",m_conn);
									data = cmd.ExecuteReader();
									if(data.HasRows){
                                                                                //Check if is already friend or has already requested friendship
                                                                                data.Close();
                                                                                cmd= new OdbcCommand("SELECT COUNT(*) FROM Friend, FriendReq WHERE (Friend.User ='"+m_user+"' AND Friend.OtherUser='"+friend+"') OR (Friend.OtherUser ='"+m_user+"' AND Friend.User='"+friend+"') OR (FriendReq.OtherUser ='"+m_user+"' AND Friend.RequestingUser='"+friend+"') OR (FriendReq.RequestingUser ='"+m_user+"' AND Friend.OtherUser='"+friend+"');",m_conn);
                                                                                data=cmd.ExecuteReader();
                                                                                if(data.HasRows){

											cmd= new OdbcCommand("INSERT INTO FriendReq VALUES ('"+m_user+"','"+friend+"');",m_conn);
											try{cmd.ExecuteNonQuery();
											rep="1";
                                                                                        }
                                                                                        catch(OdbcException oe){
                                                                                        	rep="0";
                                                                                        }
                                                                                }
                                                                                else
                                                                                	rep="0";
									}
									else
                                                                                data.Close();
										rep="0";
								}
								else
									rep="0";
								bytesSent=ASCII.GetBytes(@"<FRIEND type=""add"" result="""+rep+@"""/>"+"\r\n");
								m_clientSocket.Send(bytesSent);

								break;
							case "DELETE":

								rep=doc.DocumentElement.GetAttribute("user");
								if (rep.Equals(m_user)){
									cmd = new OdbcCommand("DELETE FROM Friend WHERE (User='"+m_user+"' AND OtherUser='"+doc.DocumentElement.GetAttribute("friend")+"') OR (User='"+doc.DocumentElement.GetAttribute("friend")+"' AND OtherUser='"+m_user+"');",m_conn);
									rep=cmd.ExecuteNonQuery().ToString();
								}
								else
									rep="0";
								bytesSent=ASCII.GetBytes(@"<FRIEND type=""delete"" result="""+rep+@"""/>"+"\r\n");
								m_clientSocket.Send(bytesSent);

							break;
							case "REQUEST":
								if(doc.DocumentElement.GetAttribute("user").Equals(m_user)){
									cmd= new OdbcCommand("SELECT OtherUser,User, Status FROM Friend, Online WHERE (User=Username OR OtherUser=Username) AND Username='"+m_user+"';",m_conn);
									data = cmd.ExecuteReader();
									string friends="<FRIENDS>";
									while(data.Read()){
										friends+=@"<USER status="""+data.GetInt32(2)+@""">";
                                                                                if(data.GetString(0).Equals(m_user))
                                                                                	friends+=data.GetString(1)+"</USER>";
                                                                                else
                                                                                	friends+=data.GetString(0)+"</USER>";
									}
                                                                        data.Close();
                                                                        cmd=new OdbcCommand("SELECT OtherUser,User FROM Friend WHERE (User='"+m_user+"' AND OtherUser NOT IN(SELECT Username FROM Online)) OR (OtherUser='"+m_user+"' AND User NOT IN(SELECT Username FROM Online));",m_conn);
                                                                        data=cmd.ExecuteReader();
                                                                        while(data.Read()){
										friends+=@"<USER status=""-1"">";
                                                                                if(data.GetString(0).Equals(m_user))
                                                                                	friends+=data.GetString(1)+"</USER>";
                                                                                else
                                                                                	friends+=data.GetString(0)+"</USER>";
									}
									friends+="</FRIENDS>\r\n";
									bytesSent=ASCII.GetBytes(friends);

								}
                                                                else{
                                                                       bytesSent=ASCII.GetBytes("<FRIENDS/>\r\n");
                                                                }
                                                                m_clientSocket.Send(bytesSent,bytesSent.Length,0);

							break;
                                                        case "ACCEPT":
                                                        	string success=@"<FRIEND type=""accept"" result=""1""/>"+"\r\n";
                                                                string failed=@"<FRIEND type=""accept"" result=""0""/>"+"\r\n";
                                                        	if(doc.DocumentElement.GetAttribute("user").Equals(m_user)){
                                                                	cmd=new OdbcCommand("SELECT FriendReqID WHERE OtherUser='"+m_user+"' AND RequestingUser='"+doc.DocumentElement.GetAttribute("other")+"';",m_conn);
                                                                        data=cmd.ExecuteReader();
                                                                        if(data.HasRows){
                                                                        	data.Read();
                                                                                int id=data.GetInt16(0);
                                                                                data.Close();
                                                                                cmd=new OdbcCommand("DELETE FROM FriendReq WHERE FriendReqId="+id.ToString(),m_conn);
                                                                                cmd.ExecuteNonQuery();
                                                                                cmd=new OdbcCommand("INSERT INTO FRIEND (User, OtherUser) VALUES ("+doc.DocumentElement.GetAttribute("other")+","+m_user+");",m_conn);
                                                                                cmd.ExecuteNonQuery();
                                                                                bytesSent=ASCII.GetBytes(success);
                                                                                m_clientSocket.Send(bytesSent);
                                                                        }
                                                                        else{
										bytesSent=ASCII.GetBytes(failed);
                                                                                m_clientSocket.Send(bytesSent);
                                                                        }
                                                                }
                                                                else{
									bytesSent=ASCII.GetBytes(failed);
                                                                        m_clientSocket.Send(bytesSent);
                                                                }
                                                                break;
                                                                case "DENY":
                                                        	string successD=@"<FRIEND type=""deny"" result=""1""/>"+"\r\n";
                                                                string failedD=@"<FRIEND type=""deny"" result=""0""/>"+"\r\n";
                                                        	if(doc.DocumentElement.GetAttribute("user").Equals(m_user)){
                                                                	cmd=new OdbcCommand("SELECT FriendReqID WHERE OtherUser='"+m_user+"' AND RequestingUser='"+doc.DocumentElement.GetAttribute("other")+"';",m_conn);
                                                                        data=cmd.ExecuteReader();
                                                                        if(data.HasRows){
                                                                        	data.Read();
                                                                                int id=data.GetInt16(0);
                                                                                data.Close();
                                                                                cmd=new OdbcCommand("DELETE FROM FriendReq WHERE FriendReqId="+id.ToString(),m_conn);
                                                                                cmd.ExecuteNonQuery();
                                                                                bytesSent=ASCII.GetBytes(successD);
                                                                                m_clientSocket.Send(bytesSent);
                                                                        }
                                                                        else{
                                                                                bytesSent=ASCII.GetBytes(failedD);
                                                                                m_clientSocket.Send(bytesSent);
                                                                        }
                                                                }
                                                                else{
                                                                        bytesSent=ASCII.GetBytes(failedD);
                                                                        m_clientSocket.Send(bytesSent);
                                                                }

                                                                break;
							}
							break;
                                                	case "ALIVE":
                                                        default:
                                                        break;
						}
                                        }
					break;
				case STATE.UNACTIVE:
					break;
				default:
					break;
			}
		}

		/// <summary>
		/// Method that checks whether there are any client calls for the
		/// last 15 seconds or not. If not this client SocketListener will
		/// be closed.
		/// </summary>
		/// <param name="o"></param>
		private void CheckClientCommInterval(object o)
		{
			if (m_lastReceiveDateTime.Equals(m_currentReceiveDateTime))
			{
				this.StopSocketListener();
			}
			else
			{
				m_lastReceiveDateTime = m_currentReceiveDateTime;
			}
		}
                private void CheckMessageBuffer(object o)
                {	ASCIIEncoding ASCII = new ASCIIEncoding();
                	if (m_user!=null){
                                lock(this) {
                        	string sql="SELECT * FROM Buffer WHERE To='"+m_user+"';";
                                OdbcConnection conn=new OdbcConnection("DSN=Message");
                                conn.Open();
				OdbcCommand cmd=new OdbcCommand(sql,conn);
		                OdbcDataReader read;
                	        byte[] bytesSent;
                       	 	read=cmd.ExecuteReader();
                                while(read.Read()){
                        		string output=@"<MESSAGE to="""+read.GetString(3)+@""" from="""+read.GetString(2)+@""">"+read.GetString(1)+"</MESSAGE>";
	                                bytesSent=ASCII.GetBytes(output);
        	                        m_clientSocket.Send(bytesSent);
        	                }
                                read.Close();
                                sql="DELETE FROM Buffer WHERE MsgTo='"+m_user+"';";
                        	OdbcCommand cmd1 = new OdbcCommand(sql,conn);
                                cmd1.ExecuteNonQuery();
                                conn.Close();
                                }
               		}

                }
                private void CheckUserOnline(object o){
                        ASCIIEncoding ASCII = new ASCIIEncoding();
                        if (m_user!=null){
                		OdbcConnection conn = new OdbcConnection("DSN=Message");
                                string sql="SELECT Online WHERE Username='"+m_user+"';";
                                conn.Open();
                                OdbcCommand cmd= new OdbcCommand(sql,conn);
                                OdbcDataReader data = cmd.ExecuteReader();
                                if(!data.HasRows){

                                        m_clientSocket.Send(ASCII.GetBytes(@"<LOGOUT user="""+m_user+@""" result=""1""/>"+"\r\n"));
                                	this.StopSocketListener();
                                }
                                data.Close();
                                conn.Close();
                        }
                }
                //Add another timer object to check if a user has requested a friendship and ask them
                private void CheckFriendInvites(object o){
                	ASCIIEncoding ASCII=new ASCIIEncoding();
                        string str="";
                        byte[] bytes;
                        if(m_user!=null){
                        	OdbcConnection conn = new OdbcConnection("DSN=Message");
                                conn.Open();
                                OdbcCommand cmd = new OdbcCommand("SELECT User FROM FriendReq WHERE OtherUser='"+m_user+"';",conn );
                                OdbcDataReader data = cmd.ExecuteReader();

                                while(data.Read()){
                                	str=@"<FRIEND type=""invite"" from="""+data.GetString(0)+@""" to="""+m_user+@"""/>"+"\r\n";
                                        bytes=ASCII.GetBytes(str);
                                        m_clientSocket.Send(bytes);
                                }
                                data.Close();
                                conn.Close();
                        }
                }
	}
}
